/* -*- buffer-read-only: t -*- vi: set ro:
   THIS FILE IS GENERATED AUTOMATICALLY.  PLEASE DO NOT EDIT.
*/
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>

static size_t debug_handle;
/* This must be defined before including debug.h */
#define MF_SOURCE_NAME  debug_handle

#include "mailfromd.h"
#include "prog.h"
#include "builtin.h"

#line 1 "io.bi"
/* This file is part of Mailfromd.             -*- c -*-
   Copyright (C) 2006, 2007, 2008, 2009, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

static size_t nstreams = 1024;

static struct mu_cfg_param io_cfg_param[] = {
	{ "max-streams", mu_cfg_size, &nstreams, 0, NULL,
	  N_("Maximum number of stream descriptors.") },
	{ NULL }
};

struct io_stream {
	char *name;
	mu_locker_t lock;
	int fd[2];
	pid_t pid;
	char *buf;
	size_t bufsize;
	void (*cleanup)(void*);
	void *cleanup_data;
};

#define IFD(s) ((s).fd[0])
#define OFD(s) ((s).fd[1] == -1 ? (s).fd[0] : (s).fd[1])

static void
flush_stream(struct io_stream *str)
{
	/*FIXME*/
}

static void
close_stream(struct io_stream *str)
{
	if (OFD(*str) == -1)
		return;
	flush_stream(str);
	close(OFD(*str));
	if (IFD(*str) != -1)
		close(IFD(*str));
	if (str->pid) {
		int status;
		waitpid(str->pid, &status, 0);
	}
	str->fd[0] = -1;
	str->fd[1] = -1;
	str->pid = 0;
	if (str->cleanup)
		str->cleanup(str->cleanup_data);
	str->cleanup = NULL;
	str->cleanup_data = NULL;
	if (str->name) {
		free(str->name);
		str->name = NULL;
	}
}

/* Read bytes from the stream STR into its buffer, until
   DELIM is encountered. Return number of bytes read. */
static int
read_stream_delim(struct io_stream *str, int delim)
{
	int fd = IFD(*str);
	size_t i = 0;
	int rc;

	for (i = 0; ; i++) {
		if (str->bufsize == i) {
			if (str->bufsize == 0) 
				str->bufsize = 16;
			str->buf = x2nrealloc(str->buf, &str->bufsize,
					      sizeof str->buf[1]);
		}
		rc = read(fd, str->buf + i, 1);
		if (rc == -1)
			return -1;
		else if (rc == 0)
			return 0;
		if ((unsigned char) str->buf[i] == delim) {
			/* NOTE: We're leaving the loop, so
			   we must increment the counter here: */
			str->buf[i++] = 0;
			break;
		}
	}
	return i;
}

#define REDIRECT_STDIN_P(f) ((f) & (O_WRONLY|O_RDWR))
#define REDIRECT_STDOUT_P(f) ((f) == O_RDONLY || ((f) & (O_RDONLY|O_RDWR)))

#define STDERR_SHUT        0
#define STDERR_NULL        1
#define STDERR_LOG         2
#define STDERR_FILE        3
#define STDERR_FILE_APPEND 4

#define LOG_TAG_PFX "mailfromd:"
#define LOG_TAG_PFX_LEN (sizeof(LOG_TAG_PFX)-1)

static void
stderr_to_log(char *arg, const char *cmd)
{
	int p[2];
	pid_t pid;
	
	if (pipe(p)) {
		mu_error(_("pipe failed: %s"), mu_strerror(errno));
		close(2);
		return;
	}

	pid = fork();

	if (pid == (pid_t) -1) {
		mu_error(_("fork failed: %s"), mu_strerror(errno));
		close(p[0]);
		close(p[1]);
		close(2);
		return;
	}
	
	/* Child */
	if (pid == 0) {
		int fd;
		FILE *fp;
		fd_set fdset;
		size_t len;
		char *buf = NULL;
		size_t bufsize = 0;
		char *tag;
		int fac = mu_log_facility, pri = LOG_ERR;
		
		if (arg) {
			char *p = strchr(arg, '.');

			if (p)
				*p++ = 0;
			if (mu_string_to_syslog_facility(arg, &fac)) {
				mu_error(_("unknown syslog facility (%s), "
					   "redirecting stderr to %s"),
					 arg,
					 mu_syslog_facility_to_string(fac));
			}
			
			if (p && mu_string_to_syslog_priority(p, &pri)) {
				mu_error(_("unknown syslog priority (%s), "
					   "redirecting stderr to %s"),
					 arg,
					 mu_syslog_priority_to_string(pri));
			}
		}
		debug2(10, "redirecting stderr to syslog %s.%s",
		       mu_syslog_facility_to_string(fac),
		       mu_syslog_priority_to_string(pri));
		
		len = strcspn(cmd, " \t");
		tag = malloc(LOG_TAG_PFX_LEN + len + 1);
		if (!tag)
			tag = (char*) cmd;
		else {
			strcpy(tag, LOG_TAG_PFX);
			memcpy(tag + LOG_TAG_PFX_LEN, cmd, len);
			tag[LOG_TAG_PFX_LEN + len] = 0;
		}
		mf_proctitle_format("%s redirector", cmd);

		FD_ZERO(&fdset);
		FD_SET(p[0], &fdset);
		fd = mflog_fd();
		if (fd != -1)
			FD_SET(fd, &fdset);
		close_fds_except(&fdset);

		fp = fdopen(p[0], "r");
#ifdef USE_SYSLOG_ASYNC		
		if (use_syslog_async)
			openlog_async(tag, LOG_PID, mu_log_facility);
		else
#else
			openlog(tag, LOG_PID, mu_log_facility);
#endif
		while (getline(&buf, &bufsize, fp) > 0)
			syslog(pri, "%s", buf);
		exit(0);
	}

	/* Parent */
	close(p[0]);
	dup2(p[1], 2);
	close(p[1]);
}
	
static void
stderr_handler(int mode, char *arg, const char *cmd)
{
	int fd;
	int append = O_TRUNC;
	
	switch (mode) {
	case STDERR_SHUT:
		close(2);
		break;

	case STDERR_NULL:
		arg = "/dev/null";
	case STDERR_FILE_APPEND:
		append = O_APPEND;
	case STDERR_FILE:
		if (!arg || !*arg) {
			close(2);
			break;
		}
		debug1(10, "redirecting stderr to %s", arg);
		fd = open(arg, O_CREAT|O_WRONLY|append, 0644);
		if (fd < 0) {
			mu_error(_("cannot open file %s for appending: %s"),
				 arg, mu_strerror(errno));
			close(2);
			return;
		}
		if (fd != 2) {
			dup2(fd, 2);
			close(fd);
		}
		break;

	case STDERR_LOG:
		stderr_to_log(arg, cmd);
	}
}

static void
parse_stderr_redirect(const char **pcmd, int *perr, char **parg)
{
	int err;
	size_t len;
	char *arg;
	const char *cmdline = *pcmd;
	
	while (*cmdline && c_isspace(*cmdline))
		cmdline++;
	if (strncmp(cmdline, "2>file:", 7) == 0) {
		cmdline += 7;
		err = STDERR_FILE;
	} else if (strncmp(cmdline, "2>>file:", 8) == 0) {
		cmdline += 8;
		err = STDERR_FILE_APPEND;
	} else if (strncmp(cmdline, "2>null:", 7) == 0) {
		cmdline += 7;
		err = STDERR_NULL;
	} else if (strncmp(cmdline, "2>syslog:", 9) == 0) {
		cmdline += 9;
		err = STDERR_LOG;
	} else
		return;

	len = strcspn(cmdline, " \t");
	if (len > 0 && cmdline[len-1] == 0)
		return;
	if (len == 0)
		arg = NULL;
	else {
		arg = malloc(len + 1);
		if (!arg)
			return;
		memcpy(arg, cmdline, len);
		arg[len] = 0;
	}

	*pcmd = cmdline + len;
	*perr = err;
	*parg = arg;
}


static int
open_program_stream(eval_environ_t env,
		    struct io_stream *str, const char *cmdline,
		    int flags)
{
	int rightp[2], leftp[2];
	int rc = 0;
	pid_t pid;
	int argc;
	char **argv;
	int err = STDERR_SHUT;
	char *arg = NULL;

	parse_stderr_redirect(&cmdline, &err, &arg);
	while (*cmdline && (*cmdline == ' ' || *cmdline == '\t'))
		cmdline++;
	
	if (REDIRECT_STDIN_P(flags))
		pipe(leftp);
	if (REDIRECT_STDOUT_P(flags))
		pipe(rightp);

	switch (pid = fork()) {
		/* The child branch.  */
	case 0:
		/* attach the pipes */

		/* Right-end */
		if (REDIRECT_STDOUT_P(flags)) {
			if (rightp[1] != 1) {
				close(1);
				dup2(rightp[1], 1);
			}
			close(rightp[0]);
		}

		/* Left-end */
		if (REDIRECT_STDIN_P(flags)) {
			if (leftp[0] != 0) {
				close(0);
				dup2(leftp[0], 0);
			}
			close(leftp[1]);
		}
      
		stderr_handler(err, arg, cmdline);
		
		/* Close unneded descripitors */
		close_fds_above(2);

		debug1(10, "running %s", cmdline);
		rc = mu_argcv_get(cmdline, "", NULL, &argc, &argv);
		if (rc) {
			mu_error(_("cannot parse command line %s: %s"),
				 cmdline, mu_strerror(rc));
			exit(127);
		}			
		execvp(argv[0], argv);
		mu_error(_("cannot run %s: %s"),
			 cmdline, mu_strerror(rc));
		exit(127);
		/********************/

		/* Parent branches: */
	case -1:
		/* Fork has failed */
		/* Restore things */
		rc = errno;
		if (REDIRECT_STDOUT_P(flags)) {
			close(rightp[0]);
			close(rightp[1]);
		}
		if (REDIRECT_STDIN_P(flags)) {
			close(leftp[0]);
			close(leftp[1]);
		}
		break;
		
	default:
		str->pid = pid;
		if (REDIRECT_STDOUT_P(flags)) {
			str->fd[0] = rightp[0];
			close(rightp[1]);
		} else
			str->fd[0] = -1;

		if (REDIRECT_STDIN_P(flags)) {
			str->fd[1] = leftp[1];
			close(leftp[0]);
		} else
			str->fd[1] = -1;
	}
	free(arg);
	return rc;
}


static int
open_file_stream(eval_environ_t env,
		 struct io_stream *str, const char *file, int flags)
{
	str->fd[0] = open(file, flags, 0644); /* FIXME: mode? */
	if (str->fd[0] == -1)
		return errno;
	return 0;
}



static int
open_parsed_inet_stream(eval_environ_t env,
			struct io_stream *str,
			const char *cstr,
			char *proto, char *port, char *path,
			int flags)
{
	union {
		struct sockaddr sa;
		struct sockaddr_in s_in;
		struct sockaddr_un s_un;
	} addr;

	socklen_t socklen;
	int fd;
	int rc;

	if (!proto
	    || strcmp(proto, "unix") == 0 || strcmp(proto, "local") == 0) {
		struct stat st;
		
			if (!(port == NULL))
#line 424
			env_throw_bi(env, mfe_failure, NULL, _("invalid connection type: %s; "
			    "port is meaningless for UNIX sockets"),cstr)
#line 428
;
		
			if (!(strlen(path) <= sizeof addr.s_un.sun_path))
#line 430
			env_throw_bi(env, mfe_range, NULL, _("%s: UNIX socket name too long"),path)
#line 433
;
		
		addr.sa.sa_family = PF_UNIX;
		socklen = sizeof(addr.s_un);
		strcpy(addr.s_un.sun_path, path);
		
		if (stat(path, &st)) {
				env_throw_bi(env, mfe_failure, NULL, _("%s: cannot stat socket: %s"),path,strerror(errno));
#line 443
		} else {
			/* FIXME: Check permissions? */
				if (!(S_ISSOCK(st.st_mode)))
#line 445
			env_throw_bi(env, mfe_failure, NULL, _("%s: not a socket"),path)
#line 448
;
		}

	} else if (strcmp(proto, "inet") == 0) {
		short pnum;
		long num;
		char *p;
		
		addr.sa.sa_family = PF_INET;
		socklen = sizeof(addr.s_in);

			if (!(port != NULL))
#line 459
			env_throw_bi(env, mfe_failure, NULL, _("invalid connection type: %s; "
			    "missing port number"),cstr)
#line 463
;

		num = pnum = strtol(port, &p, 0);
		if (*p == 0) {
				if (!(num == pnum))
#line 467
			env_throw_bi(env, mfe_range, NULL, _("invalid connection type: "
				    "%s; bad port number"),cstr)
#line 471
;
			pnum = htons(pnum);
		} else {
			struct servent *sp = getservbyname(port, "tcp");

				if (!(sp != NULL))
#line 476
			env_throw_bi(env, mfe_failure, NULL, _("invalid connection type: "
				    "%s; unknown port name"),cstr)
#line 480
;
			pnum = sp->s_port;
		}
		
		if (!path)
			addr.s_in.sin_addr.s_addr = INADDR_ANY;
		else {
			struct hostent *hp = gethostbyname(path);
				if (!(hp != NULL))
#line 488
			env_throw_bi(env, mfe_failure, NULL, _("unknown host name %s"),path)
#line 491
;
			addr.sa.sa_family = hp->h_addrtype;
			switch (hp->h_addrtype) {
			case AF_INET:
				memmove(&addr.s_in.sin_addr, hp->h_addr, 4);
				addr.s_in.sin_port = pnum;
				break;

			default:
					env_throw_bi(env, mfe_range, NULL, _("invalid connection type: "
					   "%s; unsupported address family"),cstr);
#line 504
			}
		}
	} else {
			env_throw_bi(env, mfe_range, NULL, _("unsupported protocol: %s"),proto);
#line 510
	}

	fd = socket(addr.sa.sa_family, SOCK_STREAM, 0);
		if (!(fd != -1))
#line 513
			env_throw_bi(env, mfe_failure, NULL, _("unable to create new socket: %s"),strerror(errno))
#line 516
;

	/* FIXME: Bind to the source ? */

	rc = connect(fd, &addr.sa, socklen);
	if (rc) {
		close(fd);
			env_throw_bi(env, mfe_failure, NULL, _("cannot connect to %s: %s"),cstr,strerror(errno));
#line 526
	}
	
	str->fd[0] = fd;
	return 0;
}

static int
open_inet_stream(eval_environ_t env,
		 struct io_stream *str, const char *addr, int flags)
{
	int rc;
	char *proto, *port, *path;

	if (gacopyz_parse_connection(addr, &proto, &port, &path)
	    != MI_SUCCESS) 
		rc = ENOMEM; /* FIXME: or EINVAL? */
	else {
		rc = open_parsed_inet_stream(env,
					     str, addr,
					     proto, port, path, flags);
		free(proto);
		free(port);
		free(path);
	}
	return rc;
}


static void *
alloc_streams()
{
	struct io_stream *p, *stab = xcalloc(nstreams, sizeof *stab);
	for (p = stab; p < stab + nstreams; p++) 
		p->fd[0] = p->fd[1] = -1;
	return stab;
}

static void
destroy_streams(void *data)
{
	struct io_stream *stab = data;
	struct io_stream *p;
	for (p = stab; p < stab + nstreams; p++) {
		close_stream(p);
		free(p->buf);
	}
	free(stab);
}
		

#line 575

#line 575
static int IO_id;
#line 575 "io.bi"



void
#line 578
bi_open(eval_environ_t env)
#line 578

#line 578 "io.bi"
{
#line 578
	
#line 578

#line 578

#line 578
char * name;
#line 578
        
#line 578
get_string_arg(env, 0, &name);
#line 578
        
#line 578

#line 578
        adjust_stack(env, 1);
#line 578

#line 578
	if (builtin_module_trace(BUILTIN_IDX_io))
#line 578
		prog_trace(env, "open %s",name);;
#line 578

{
	int i, rc;
	int flags = 0;
	int (*opf)(eval_environ_t env,
		   struct io_stream *, const char *, int) = open_file_stream;
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	
	for (i = 0; i < nstreams; i++) {
		if (iotab[i].fd[0] == -1) 
			break;
	}
		if (!(i < nstreams))
#line 590
			env_throw_bi(env, mfe_failure, "open", _("no more files available"))
#line 592
;

	debug1(10, "opening stream %s", name);
	iotab[i].name = xstrdup(name);
	if (*name == '>') {
		flags |= O_RDWR|O_CREAT;
		name++;
		if (*name == '>') {
			flags |= O_APPEND;
			name++;
		} else
			flags |= O_TRUNC;
	} else if (*name == '|') {
		opf = open_program_stream;
		flags = O_WRONLY;
		name++;
		if (*name == '&') {
			flags = O_RDWR;
			name++;
		}
	} else if (*name == '@') {
		name++;
		opf = open_inet_stream;
		flags = O_RDWR;
	} else
		flags = O_RDONLY;
	
	for (;*name && c_isspace(*name); name++)
		;
	
	rc = opf(env, &iotab[i], name, flags);
	
		if (!(rc == 0))
#line 624
			env_throw_bi(env, mfe_failure, "open", _("cannot open stream %s: %s"),name,mu_strerror(rc))
#line 627
;
	debug2(10, "open(%s) = %d", name, i);
	
#line 629
do {
#line 629
  push(env, (STKVAL) (i));
#line 629
  return;
#line 629
} while (0);
}
}

void
#line 633
bi_close(eval_environ_t env)
#line 633

#line 633 "io.bi"
{
#line 633
	
#line 633

#line 633

#line 633
long  fd;
#line 633
        
#line 633
get_numeric_arg(env, 0, &fd);
#line 633
        
#line 633

#line 633
        adjust_stack(env, 1);
#line 633

#line 633
	if (builtin_module_trace(BUILTIN_IDX_io))
#line 633
		prog_trace(env, "close %lu",fd);;
#line 633

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);

		if (!(fd >= 0 && fd < nstreams))
#line 637
			env_throw_bi(env, mfe_range, "close", _("invalid file descriptor"))
#line 639
;
	close_stream(&iotab[fd]);
}
}	

void
#line 644
bi_write(eval_environ_t env)
#line 644

#line 644 "io.bi"
{
#line 644
	
#line 644

#line 644
long __bi_argcnt;
#line 644
long  fd;
#line 644
        char * str;
#line 644
        long  n;
#line 644
        
#line 644
get_numeric_arg(env, 1, &fd);
#line 644
        get_string_arg(env, 2, &str);
#line 644
        get_numeric_arg(env, 3, &n);
#line 644
        
#line 644
get_numeric_arg(env, 0, &__bi_argcnt);
#line 644
        adjust_stack(env, __bi_argcnt + 1);
#line 644

#line 644
	if (builtin_module_trace(BUILTIN_IDX_io))
#line 644
		prog_trace(env, "write %lu %s %lu",fd, str, ((__bi_argcnt > 2) ? n : 0));;
#line 644

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	int rc;
	
	debug2(10, "writing %s to %d", str, fd);
		if (!(fd >= 0 && fd < nstreams && OFD(iotab[fd])))
#line 650
			env_throw_bi(env, mfe_range, "write", _("invalid file descriptor"))
#line 652
;
	if (!(__bi_argcnt > 2))
		n = strlen (str);
	rc = write(OFD(iotab[fd]), str, n);
		if (!(n == rc))
#line 656
			env_throw_bi(env, mfe_io, "write", _("write error on %s: %s"),iotab[fd].name,mu_strerror(errno))
#line 659
;
}
}

void
#line 663
bi_read(eval_environ_t env)
#line 663

#line 663 "io.bi"
{
#line 663
	
#line 663

#line 663

#line 663
long  fd;
#line 663
        long  size;
#line 663
        
#line 663
get_numeric_arg(env, 0, &fd);
#line 663
        get_numeric_arg(env, 1, &size);
#line 663
        
#line 663

#line 663
        adjust_stack(env, 2);
#line 663

#line 663
	if (builtin_module_trace(BUILTIN_IDX_io))
#line 663
		prog_trace(env, "read %lu %lu",fd, size);;
#line 663

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	int rc;
	size_t off;
	char *s = (char*) env_data_ref(env, (off = heap_reserve(env, size + 1)));

		if (!(fd >= 0 && fd < nstreams && IFD(iotab[fd])))
#line 670
			env_throw_bi(env, mfe_range, "read", _("invalid file descriptor"))
#line 672
;
	
	rc = read(IFD(iotab[fd]), s, size);
	if (rc == 0)
			env_throw_bi(env, mfe_eof, "read", _("EOF on %s"),iotab[fd].name);
#line 678
		if (!(rc == size))
#line 678
			env_throw_bi(env, mfe_io, "read", _("read error on %s: %s"),iotab[fd].name,mu_strerror(errno))
#line 681
;
	s[size] = 0;
	
#line 683
do {
#line 683
  push(env, (STKVAL) (off));
#line 683
  return;
#line 683
} while (0);
}	
}

/* FIXME: only first char of delim is used */
void
#line 688
bi_getdelim(eval_environ_t env)
#line 688

#line 688 "io.bi"
{
#line 688
	
#line 688

#line 688

#line 688
long  fd;
#line 688
        char * delim;
#line 688
        
#line 688
get_numeric_arg(env, 0, &fd);
#line 688
        get_string_arg(env, 1, &delim);
#line 688
        
#line 688

#line 688
        adjust_stack(env, 2);
#line 688

#line 688
	if (builtin_module_trace(BUILTIN_IDX_io))
#line 688
		prog_trace(env, "getdelim %lu %s",fd, delim);;
#line 688

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	int rc;

		if (!(fd >= 0 && fd < nstreams && IFD(iotab[fd])))
#line 693
			env_throw_bi(env, mfe_range, "getdelim", _("invalid file descriptor"))
#line 695
;
	rc = read_stream_delim(&iotab[fd], delim[0]); 
	if (rc == 0)
			env_throw_bi(env, mfe_eof, "getdelim", _("EOF on %s"),iotab[fd].name);
#line 700
		if (!(rc > 0))
#line 700
			env_throw_bi(env, mfe_io, "getdelim", _("read error on %s: %s"),iotab[fd].name,mu_strerror(errno))
#line 703
;
	do {pushs(env, iotab[fd].buf);
#line 704
return;
#line 704
} while (0);
}	
}

void
#line 708
bi_getline(eval_environ_t env)
#line 708

#line 708 "io.bi"
{
#line 708
	
#line 708

#line 708

#line 708
long  fd;
#line 708
        
#line 708
get_numeric_arg(env, 0, &fd);
#line 708
        
#line 708

#line 708
        adjust_stack(env, 1);
#line 708

#line 708
	if (builtin_module_trace(BUILTIN_IDX_io))
#line 708
		prog_trace(env, "getline %lu",fd);;
#line 708

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	int rc;

		if (!(fd >= 0 && fd < nstreams && IFD(iotab[fd])))
#line 713
			env_throw_bi(env, mfe_range, "getline", _("invalid file descriptor"))
#line 715
;
	rc = read_stream_delim(&iotab[fd], '\n');
	if (rc == 0)
			env_throw_bi(env, mfe_eof, "getline", _("EOF on %s"),iotab[fd].name);
#line 720
		if (!(rc > 0))
#line 720
			env_throw_bi(env, mfe_io, "getline", _("read error on %s: %s"),iotab[fd].name,mu_strerror(errno))
#line 723
;
	do {pushs(env, iotab[fd].buf);
#line 724
return;
#line 724
} while (0);
}	
}


#line 728
void
#line 728
io_init_builtin()
#line 728
{
#line 728
        debug_handle = debug_register("bi_io");
#line 728
	#line 575 "io.bi"
IO_id = builtin_priv_register(alloc_streams, destroy_streams,
#line 575
NULL);
#line 578 "io.bi"
va_builtin_install_ex("open", bi_open, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 633 "io.bi"
va_builtin_install_ex("close", bi_close, 0, dtype_unspecified, 1, 0, 0|0, dtype_number);
#line 644 "io.bi"
va_builtin_install_ex("write", bi_write, 0, dtype_unspecified, 3, 1, 0|0, dtype_number, dtype_string, dtype_number);
#line 663 "io.bi"
va_builtin_install_ex("read", bi_read, 0, dtype_string, 2, 0, 0|0, dtype_number, dtype_number);
#line 688 "io.bi"
va_builtin_install_ex("getdelim", bi_getdelim, 0, dtype_string, 2, 0, 0|0, dtype_number, dtype_string);
#line 708 "io.bi"
va_builtin_install_ex("getline", bi_getline, 0, dtype_string, 1, 0, 0|0, dtype_number);

#line 728 "io.bi"
	
#line 728
	 mf_add_runtime_params(io_cfg_param);
#line 728
	 
#line 728
}
#line 730

#line 627 "../../mfd/builtin/snarf.m4"

#line 627

